/* << human ps >>
 * hps.c       ver. 1.01
 * 96/02/03
 * By Kenta Cho(cho@ueda.info.waseda.ac.jp)
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "hps.h"

#define MAXLINE 255

typedef struct {
  int pid,pcpu;
  char pname[21],puser[9];
} Ps;

typedef struct {
  Ps ps;
  int ncpu;
  int xp,yp;
} Sps;

Sps sps[MAXPSMAX];
int maxps=4;
int xpsnum=-1;
char psstr[256]={PSCOMMAND};

XFontStruct *fo;
int f_height;
char fontname[256]={FONTNAME};

Display *disp;
Window win;
GC gc;

char windowtitle[256]={"hps"};
int xscsize,yscsize;
int xwinsize,ywinsize;
int xwinoffset,ywinoffset;

int screeninterval=SCREENINTERVAL,psinterval=PSINTERVAL;

int fsnapshot=SNAPSHOT;

char geom[256];

char dpy[256]={NULL};

int pidstrpo,cpustrpo,comstrpo;

int fplt=1,fpco=0;

int manpoint[][2]=
  {{0,96},{-18,86},{-18,75},{0,64},{18,75},{18,86},{0,96},{-1,0},
     {0,44},{-24,64},{-24,96},{-1,0},
     {0,44},{24,64},{24,96},{-1,0},
     {0,64},{0,32},{-24,0},{-1,0},
     {0,32},{24,0},{-1,0},{-2,0}};

/*
 * initializing
 */
initall()
{
  int r,i,c=0;
  int ypsnum;
  for ( r=0 ; r<maxps ; r++ ) sps[r].ncpu=-1;
  ypsnum=(int)((maxps-1)/xpsnum)+1;
  xscsize=256*xpsnum;
  yscsize=256*ypsnum;

  xwinsize=xscsize*SCREENSIZE;
  ywinsize=yscsize*SCREENSIZE;
  for ( r=0 ; r<ypsnum ; r++ )
    for ( i=0 ; i<xpsnum ; i++ ) {
      sps[c].ncpu=-1;
      sps[c].xp=i*256;
      sps[c].yp=r*256;
      c++;
    }
}

/*
 * display usage
 */
usage()
{
  fprintf(stderr,"Usage:  hps [-options ...]\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"where options include:\n");
  fprintf(stderr,"    -display dpy            X server on which to display\n");
  fprintf(stderr,"    -geometry geom          size and location of window\n");
  fprintf(stderr,"    -fn font                font to use\n");
  fprintf(stderr,"    -si 0.1seconds          interval between screen updates\n");
  fprintf(stderr,"    -pi 0.1seconds          interval between ps updates\n");
  fprintf(stderr,"    -mn number              maximum number of processes\n");
  fprintf(stderr,"    -xn number              number of horizontal processes\n");
  fprintf(stderr,"    -an                     show animation\n");
  fprintf(stderr,"    -sn                     show snapshot\n");
}

/*
 * parse options
 */
parseoptions(int argc,char **argv)
{
  int r;
  for ( r=1 ; r<argc ; r++ ) {
    if ( strcmp(argv[r],"-si")==0 && r<argc-1 ) {
      screeninterval=atoi(argv[r+1]); r++;
      fsnapshot=0;
    }
    else if ( strcmp(argv[r],"-pi")==0 && r<argc-1 ) {
      psinterval=atoi(argv[r+1]); r++;
      fsnapshot=0;
    }
    else if ( strcmp(argv[r],"-mn")==0 && r<argc-1 ) {
      maxps=atoi(argv[r+1]); r++;
    }
    else if ( strcmp(argv[r],"-xn")==0 && r<argc-1 ) {
      xpsnum=atoi(argv[r+1]); r++;
    }
    else if ( strcmp(argv[r],"-an")==0 ) fsnapshot=0;
    else if ( strcmp(argv[r],"-sn")==0 ) fsnapshot=1;
    else if ( strcmp(argv[r],"-fn")==0 && r<argc-1 ) {
      strcpy(fontname,argv[r+1]); r++;
    }
    else if ( (strcmp(argv[r],"-display")==0 || strcmp(argv[r],"-d")==0) && r<argc-1 ) {
      strcpy(dpy,argv[r+1]); r++;
    }
    else if ( (strcmp(argv[r],"-geometry")==0 || strcmp(argv[r],"-g")==0) && r<argc-1 ) {
      strcpy(geom,argv[r+1]);
      r++;
    }
    else {
      usage(); exit(1);
    }
  }
  if ( maxps>MAXPSMAX ) maxps=MAXPSMAX;
  if ( xpsnum<0 || xpsnum>maxps ) xpsnum=maxps;
}

/*
 * open window
 */
openwindow(char *title)
{
  XSizeHints xsh;
  int f;

  if ( NULL==(disp=XOpenDisplay(dpy)) ) {
    fprintf(stderr,"Error: I cannot open display: %s\n",dpy);
    exit(0);
  }

  if ( NULL==(fo=XLoadQueryFont(disp,fontname)) ) {
    fprintf(stderr,"Error: I cannot convert string \"%s\" to tyoe FontStruct\n",fontname);
    exit(0);
  }

  f=XWMGeometry(disp,DefaultScreen(disp),geom,NULL,1,&xsh,&(xsh.x),&(xsh.y),&(xsh.width),&(xsh.height),&(xsh.win_gravity));
  xsh.flags=0;
  if ( (f&(XValue | YValue))!=0 ) {
    xsh.flags|=USPosition;
    xwinoffset=xsh.x; ywinoffset=xsh.y;
  }
  if ( (f&(WidthValue | HeightValue))!=0 ) {
    xsh.flags|=USSize;
    xwinsize=xsh.width; ywinsize=xsh.height;
  }

  win=XCreateSimpleWindow(disp,RootWindow(disp,0),xwinoffset,ywinoffset,xwinsize,ywinsize,1,WhitePixel(disp,0),BlackPixel(disp,0));
  XStoreName(disp,win,title);
  XSetWMNormalHints(disp,win,&xsh);
  XMapWindow(disp,win);
  XSelectInput(disp,win,StructureNotifyMask | ExposureMask);
  gc=XCreateGC(disp,win,0,NULL);
  XSetForeground(disp,gc,WhitePixel(disp,0));

  f_height=fo->max_bounds.ascent+fo->max_bounds.descent;
  XSetFont(disp,gc,fo->fid); 
}

/*
 * draw line
 */
setline(int x1,int y1,int x2,int y2)
{
  int sx1,sy1,sx2,sy2;
  sx1=x1*xwinsize/xscsize;
  sx2=x2*xwinsize/xscsize;
  sy1=y1*ywinsize/yscsize;
  sy2=y2*ywinsize/yscsize;
  XDrawLine(disp,win,gc,sx1,sy1,sx2,sy2);
}

/*
 * clear area
 */
cleararea(int x1,int y1,int wid,int hei)
{
  int sx1,sy1,swid,shei;
  sx1=x1*xwinsize/xscsize;
  swid=wid*xwinsize/xscsize;
  sy1=y1*ywinsize/yscsize;
  shei=hei*ywinsize/yscsize;
  XClearArea(disp,win,sx1,sy1,swid,shei,0);
}

/*
 * draw string
 */
drawstring(int x,int y,char *str)
{
  int sx,sy;
  sx=x*xwinsize/xscsize;
  sy=y*ywinsize/yscsize;
  XDrawString(disp,win,gc,sx,sy,str,strlen(str));
}

/*
 * display human
 */
putman(int x,int y,int cpu,char *pname,char *uname)
{
  int r=0;
  int sx,sy,nsx,nsy,oy,fx,fy;
  char cpustr[8];

  oy=-manpoint[r][1]*(1000-cpu)/1000+y;
  while ( manpoint[r][0]!=-2 ) {
    sx=manpoint[r][0]+x; sy=-manpoint[r][1]*(1000-cpu)/1000+y;
    r++;
    while ( manpoint[r][0]!=-1 ) {
      nsx=manpoint[r][0]+x; nsy=-manpoint[r][1]*(1000-cpu)/1000+y;
      setline(sx,sy,nsx,nsy);
      sx=nsx; sy=nsy;
      r++;
    }
    r++;
  }

  fx=48+(64*cpu)/1000;
  fy=32+(128*cpu)/1000;
  setline(x-fx,oy,x+fx,oy);
  setline(x+fx,oy,x+fx*0.6,oy-fy);
  setline(x+fx*0.6,oy-fy,x+fx*0.2,oy-fy);
  setline(x+fx*0.2,oy-fy,x+fx*0.2,oy-fy*1.2);
  setline(x+fx*0.2,oy-fy*1.2,x-fx*0.2,oy-fy*1.2);
  setline(x-fx*0.2,oy-fy*1.2,x-fx*0.2,oy-fy);
  setline(x-fx*0.2,oy-fy,x-fx*0.6,oy-fy);
  setline(x-fx*0.6,oy-fy,x-fx,oy);
  sprintf(cpustr,"%2.1f%%",(float)cpu/10);
  drawstring(x-48,oy,cpustr);
  drawstring(x-100,oy-fy,pname);
  drawstring(x-64,240,uname);
}

/*
 * display processes
 */
putps(int f)
{
  int r,i;
  int bcpu;

  for ( r=0 ; r<maxps ; r++ ) {
    if ( sps[r].ncpu>=0 ) {
      if ( fsnapshot ) {
	sps[r].ncpu=sps[r].ps.pcpu;
	putman(sps[r].xp+128,sps[r].yp+224,sps[r].ncpu,sps[r].ps.pname,sps[r].ps.puser);
      }
      else {
	bcpu=sps[r].ncpu;
	for ( i=0 ; i<screeninterval ; i++ ) {
	  sps[r].ncpu+=(((sps[r].ps.pcpu-sps[r].ncpu)+16)>>5);
	  if ( sps[r].ncpu>sps[r].ps.pcpu ) sps[r].ncpu--;
	  if ( sps[r].ncpu<sps[r].ps.pcpu ) sps[r].ncpu++;
	}
	
	if ( sps[r].ncpu<=0 ) sps[r].ncpu=-1;
	if ( bcpu!=sps[r].ncpu || f ) {
	  if ( bcpu>0 ) cleararea(sps[r].xp,sps[r].yp,256,256);
	  if ( sps[r].ncpu>0 ) putman(sps[r].xp+128,sps[r].yp+224,sps[r].ncpu,sps[r].ps.pname,sps[r].ps.puser);
	}
      }
    }
  }
}

/*
 * check ps
 */
checkps()
{
  FILE *fp;
  char line[MAXLINE];
  int r;

  if ( (fp=popen(psstr,"r"))==NULL ) {
    fprintf(stderr,"Error: I cannot use 'ps' command\n");
    exit(0);
  }
  fgets(line,MAXLINE,fp);
  pclose(fp);
  for ( r=0 ; r<strlen(line) ; r++ ) {
    if ( strncmp(&line[r],"PID",3)==0 ) {
      pidstrpo=r-2;
      for ( ; r<strlen(line) ; r++ ) {
	if ( strncmp(&line[r],"%CPU",4)==0 ) {
	  cpustrpo=r;
	  for ( ; r<strlen(line) ; r++ ) {
	    if ( strncmp(&line[r],"COMMAND",7)==0 ) {
	      comstrpo=r;
	      return;
	    }
	  }
	}
      }
    }
  }
  fprintf(stderr,"Error: I cannot recognize 'ps' command\n");
  exit(0);
}

/*
 * get the status of processes
 */
int getps(Ps ps[])
{
  FILE *fp;
  char line[MAXLINE],ncom[MAXLINE];
  float cpu;
  int r;

  if ( (fp=popen(psstr,"r"))==NULL ) {
    fprintf(stderr,"Warning: I cannot use 'ps' command now\n");
    return;
  }
  fgets(line,MAXLINE,fp);
  for ( r=0 ; r<maxps ; r++ ) {
    if ( feof(fp) ) break;
    fgets(line,MAXLINE,fp);
    sscanf(&line[0],"%8s",ps[r].puser);
    sscanf(&line[pidstrpo],"%d",&ps[r].pid);
    sscanf(&line[cpustrpo],"%f",&cpu);
    sscanf(&line[comstrpo],"%12s",ps[r].pname);
    if ( (ps[r].pcpu=(int)(cpu*10))==0 ) break;
  }
  pclose(fp);
  return(r);
}
  
/*
 * update process data
 */
updatesps()
{
  Ps ps[MAXPSMAX];
  int nps;
  int r,i;
  int ef[MAXPSMAX],nf[MAXPSMAX];

  nps=getps(ps);
  
  for ( r=0 ; r<maxps ; r++ ) ef[r]=nf[r]=0;
  for ( r=0 ; r<maxps ; r++ ) {
    if ( sps[r].ncpu>=0 ) {
      for ( i=0 ; i<nps ; i++ ) {
	if ( sps[r].ps.pid==ps[i].pid ) {
	  ef[r]=1; nf[i]=1;
	  sps[r].ps.pcpu=ps[i].pcpu;
	  break;
	}
      }
    }
  }
  r=0;
  for ( i=0 ; i<nps ; i++ ) {
    if ( nf[i]==0 ) {
      for ( ; r<maxps ; r++ ) {
	if ( sps[r].ncpu<0 ) {
	  sps[r].ps.pid=ps[i].pid;
	  sps[r].ps.pcpu=ps[i].pcpu;
	  strcpy(sps[r].ps.pname,ps[i].pname);
	  strcpy(sps[r].ps.puser,ps[i].puser);
	  sps[r].ncpu=0;
	  ef[r]=1; nf[i]=1;
	  break;
	}
      }
    }
  }
  r=0;
  for ( i=0 ; i<nps ; i++ ) {
    if ( nf[i]==0 ) {
      for ( ; r<maxps ; r++ ) {
	if ( ef[r]==0 ) {
	  sps[r].ps.pid=ps[i].pid;
	  sps[r].ps.pcpu=ps[i].pcpu;
	  strcpy(sps[r].ps.pname,ps[i].pname);
	  strcpy(sps[r].ps.puser,ps[i].puser);
	  sps[r].ncpu=0;
	  ef[r]=1;
	  break;
	}
      }
    }
  }
  for ( r=0 ; r<maxps ; r++ )
    if ( ef[r]==0 && sps[r].ncpu>=0 ) sps[r].ps.pcpu=0;
}

/*
 * main routine
 */
main(int argc,char **argv)
{
  int usn;
  int r;
  int time;
  XEvent ev;

  strcat(psstr," ");
  strcat(psstr,PSOPTION);
  parseoptions(argc,argv);
  initall();
  openwindow(windowtitle);
  checkps();

  if ( fsnapshot ) {
    updatesps();
    while (1) {
      XNextEvent(disp,&ev);
      if ( ev.type==ConfigureNotify ) {
	xwinsize=ev.xconfigure.width;
	ywinsize=ev.xconfigure.height;
      }
      else if ( ev.type==Expose ) {
	XClearWindow(disp,win);
	putps(1);
      }
      XFlush(disp);
    }
  }
  else {
    while (1) {
      time=0;
      XClearWindow(disp,win);
      updatesps();
      while (1) {
	while ( XEventsQueued(disp,QueuedAfterFlush)>0 ) {
	  XNextEvent(disp,&ev);
	  if ( ev.type==ConfigureNotify ) {
	    xwinsize=ev.xconfigure.width;
	    ywinsize=ev.xconfigure.height;
	  }
	  else if ( ev.type==Expose ) {
	    XClearWindow(disp,win);
	    putps(1);
	  }
	}
	putps(0);
	XFlush(disp);
	usleep(100000*screeninterval);
	time+=screeninterval;
	if ( time>=psinterval ) break;
      }
    }
  }
}
